#ifndef _ECTOR_GENERIC_BUFFER_EO_H_
#define _ECTOR_GENERIC_BUFFER_EO_H_

#ifndef _ECTOR_GENERIC_BUFFER_EO_CLASS_TYPE
#define _ECTOR_GENERIC_BUFFER_EO_CLASS_TYPE

typedef Eo Ector_Generic_Buffer;

#endif

#ifndef _ECTOR_GENERIC_BUFFER_EO_TYPES
#define _ECTOR_GENERIC_BUFFER_EO_TYPES

typedef enum
{
  ECTOR_BUFFER_FLAG_NONE = 0, /** Buffer may not have any backing, indicates an
                               * invalid buffer. */
  ECTOR_BUFFER_FLAG_CPU_READABLE = 1, /** Can be read from the CPU after map.
                                       * Reading may still be very slow. */
  ECTOR_BUFFER_FLAG_CPU_WRITABLE = 2, /** Can be written to by the CPU after
                                       * map. Writing may still be very slow.
                                       */
  ECTOR_BUFFER_FLAG_RENDERABLE = 4, /** Can be rendered to, ie CPU memory for
                                     * SW rendering, or an FBO for GL engine.
                                     */
  ECTOR_BUFFER_FLAG_DRAWABLE = 8, /** Can be used as a source of pixels to draw
                                   * on Evas. */
  ECTOR_BUFFER_FLAG_CPU_READABLE_FAST = 16, /** Can be read by the CPU at high
                                             * speed, ie no need for
                                             * glReadPixels. */
  ECTOR_BUFFER_FLAG_CPU_WRITABLE_FAST = 32, /** Can be written by the CPU at
                                             * high speed, ie no need for GPU
                                             * texture upload. */
  ECTOR_BUFFER_FLAG_UNCACHED = 64 /** Backed by uncached memory, ie. slow-ish
                                   * reads but faster than glReadPixels. */
} Ector_Buffer_Flag;

typedef enum
{
  ECTOR_BUFFER_ACCESS_FLAG_NONE = 0,
  ECTOR_BUFFER_ACCESS_FLAG_READ = 1,
  ECTOR_BUFFER_ACCESS_FLAG_WRITE = 2,
  ECTOR_BUFFER_ACCESS_FLAG_COW = 4 /** Forces copy-on-write if already mapped
                                    * as read-only. Requires write. */
} Ector_Buffer_Access_Flag;


#endif
/** 2D pixel buffer interface for Ector
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
#define ECTOR_GENERIC_BUFFER_MIXIN ector_generic_buffer_mixin_get()

EAPI const Eo_Class *ector_generic_buffer_mixin_get(void) EINA_CONST;

/**
 * @brief Set the source pixels for this buffer, or allocate a new memory
 * region
 *
 * @param[in] width
 * @param[in] height
 * @param[in] stride Can be 0
 * @param[in] cspace
 * @param[in] writable
 * @param[in] l Left border pixels, usually 0 or 1
 * @param[in] r Right border pixels, usually 0 or 1
 * @param[in] t Top border pixels, usually 0 or 1
 * @param[in] b Bottom border pixels, usually 0 or 1
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
EOAPI Eina_Bool ector_buffer_pixels_set(void *pixels, int width, int height, int stride, Efl_Gfx_Colorspace cspace, Eina_Bool writable, unsigned char l, unsigned char r, unsigned char t, unsigned char b);

/**
 * @brief Get a single horizontal span of length w starting from (x,y)
 *
 * Call span_free() to release it. This function will try not to allocate any
 * new buffer, whenever possible. This means the data might be mapped directly
 * from the backing memory buffer.
 *
 * @param[in] y Ranges from -t to h+b-1
 * @param[in] w Ranges from 1 to w+l+r
 * @param[in] cspace Requested colorspace, may trigger conversion on the fly.
 * @param[out] length Length in bytes of the returned buffer
 *
 * @return A temporary memory buffer containing the pixels requested.
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
EOAPI uint8_t *ector_buffer_span_get(int x, int y, unsigned int w, Efl_Gfx_Colorspace cspace, unsigned int *length);

/**
 * @brief Must be called as soon as possible after span_get
 *
 * @param[in] data
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
EOAPI void ector_buffer_span_free(uint8_t *data);

/**
 * @brief Map a region of this buffer for read or write access by the CPU,
 * fetch data from the GPU if needed. This operation may be slow if
 * cpu_readable_fast or cpu_writeable_fast are not true, or if the required
 * colorspace is different from the internal one.
 *
 * @param[in] mode Specifies whether to map for read-only, write-only or
 * read-write access (OR combinaison of flags).
 * @param[in] x X position of the top-left pixel to map
 * @param[in] y Y position of the top-left pixel to map
 * @param[in] w If 0, defaults to the buffer width
 * @param[in] h If 0, defaults to the buffer height
 * @param[in] cspace Requested colorspace. If differen from the internal
 * cspace, map should try to convert the data into a new buffer
 * @param[out] stride Returns the length in bytes of a mapped line
 *
 * @return Pointer to the top-left pixel data. Returns @c null in case of
 * failure
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
EOAPI void *ector_buffer_map(unsigned int *length, Ector_Buffer_Access_Flag mode, unsigned int x, unsigned int y, unsigned int w, unsigned int h, Efl_Gfx_Colorspace cspace, unsigned int *stride);

/**
 * @brief Unmap a region of this buffer, and upload data to the GPU (if
 * needed).
 *
 * @param[in] length Must be the same as returned by map.
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
EOAPI void ector_buffer_unmap(void *data, unsigned int length);

/**
 * @brief Retrieves the (rectangular) size of the pixel buffer.
 *
 * @param[out] w
 * @param[out] h
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
EOAPI void ector_buffer_size_get(int *w, int *h);

EOAPI Efl_Gfx_Colorspace ector_buffer_cspace_get(void);

/**
 * @brief Get the capabilities of this buffer
 *
 * @return A bitmask of capability flags
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
EOAPI Ector_Buffer_Flag ector_buffer_flags_get(void);

/**
 * @brief Duplicated pixel borders of this buffer, used for GL scaling
 *
 * @param[out] l
 * @param[out] r
 * @param[out] t
 * @param[out] b
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
EOAPI void ector_buffer_border_get(int *l, int *r, int *t, int *b);

EOAPI extern const Eo_Event_Description _ECTOR_GENERIC_BUFFER_EVENT_DETACHED;

/** Emitted whenever the previously attached pixels are detached during
 * pixels_set
 *
 * @since 1.17
 *
 * @ingroup Ector_Generic_Buffer
 */
#define ECTOR_GENERIC_BUFFER_EVENT_DETACHED (&(_ECTOR_GENERIC_BUFFER_EVENT_DETACHED))

#endif
